﻿Imports System.Xml.Serialization
Public Class DownloadJob
    Public Delegate Sub dsHandleException(ByVal exc As Exception)
    Event ContinueResume(ByRef prevRangeJob As ThreadRangeJob)
    Event JobStatusChanged(ByVal newStatus As Status, ByRef dJob As DownloadJob)
    <XmlIgnore()> Friend HandleException As dsHandleException
    Public DownloadJobStatus As Status
    Dim ThreadStates() As InfoLog.ThreadState
#Region "GlobalDimensions"
    Public JobSavePath As String
    Public iLength As Int64
    Public iPort As Int64
    Public sDownloadURL As String
    Public sFileLocation As String
    Public sCookies As String
    Public sReferer As String
    Public nThreads As Int64
    Public iConnectTimeout As Int64
    Public iRecieveTimeout As Int64
    Public iSendTimeout As Int64
    Public iConnectRetries As Int64
    Public iSendRetries As Int64
    Public iRecieveRetries As Int64
    Public cItemLog As InfoLog()
    Public cStatBoard As StatBoard
    Public ThreadJobs As ThreadRangeJob()
    Public iBufferSize As Int64
    Public iThreadRetryTime As Int64
    Public canResume As Boolean
    Public isChunked As Boolean
    <XmlIgnore()> Public clsDownloader As New Downloader
#End Region
    Public jGUID As Guid
    Enum Status
        Queue = 0
        Starting = 1
        Downloading = 2
        Done = 3
        Stopped = 4
        Paused = 5
    End Enum
    Public Sub Delete()
        If IO.File.Exists(JobSavePath.PathEt & jGUID.ToString & ".xml") Then IO.File.Delete(JobSavePath.PathEt & jGUID.ToString & ".xml")
    End Sub
    Public Sub Pause()
        Dim vT As Threading.Thread
        Dim vS As Net.Sockets.Socket
        For i = 0 To clsDownloader.DownloadThreads.Length - 1
            cItemLog(i).UpdateThreadStatus(InfoLog.ThreadState.Paused)
            cItemLog(i).Update(InfoLogItem.Direction._Error, "User Cancelled. Thread Terminating.")
            vT = clsDownloader.DownloadThreads(i)
            If Not vT Is Nothing Then
                vT.Abort()
                vT = Nothing
            End If
            vS = clsDownloader.DownloadSockets(i)
            If Not vS Is Nothing Then
                If vS.Connected = True Then
                    vS.Disconnect(False)
                End If
                vS.Close()
            End If
        Next
        Save()
    End Sub
    Private Sub Save()
        If IO.Directory.Exists(JobSavePath) = False Then IO.Directory.CreateDirectory(JobSavePath)
        Dim objStreamWriter As New IO.StreamWriter(JobSavePath.PathEt & jGUID.ToString & ".xml")
        Dim x As New System.Xml.Serialization.XmlSerializer(Me.GetType)
        x.Serialize(objStreamWriter, Me)
        objStreamWriter.Close()
    End Sub

    Public Sub Start()
        clsDownloader.StartNextThread(Nothing)
    End Sub

    Private Sub ThreadStatusChanged(ByVal Status As InfoLog.ThreadState, ByRef iLog As InfoLog)
        Array.Resize(ThreadStates, cItemLog.Length)
        ThreadStates(Array.IndexOf(cItemLog, iLog)) = Status
        UpdateMainStatus(EvaluateThreadStates(ThreadStates))
    End Sub
    Private Sub ThreadStatusRecheck()
        Array.Resize(ThreadStates, cItemLog.Length)
        For i = 0 To cItemLog.Length - 1
            ThreadStates(i) = cItemLog(i).ThreadStatus
        Next
        UpdateMainStatus(EvaluateThreadStates(ThreadStates))
    End Sub
    Private Function EvaluateThreadStates(ByVal tStates() As InfoLog.ThreadState) As Status
        If tStates.Contains(InfoLog.ThreadState.Downloading) Then Return Status.Downloading
        If tStates(0) = InfoLog.ThreadState.Paused Then Return Status.Paused
        If tStates(0) = InfoLog.ThreadState.Stopped Then Return Status.Stopped
        If tStates(0) = InfoLog.ThreadState.Queue Then Return Status.Queue
        If tStates(0) = InfoLog.ThreadState.Starting Then Return Status.Starting
        Return Status.Done
    End Function
    Friend Sub MakeThreadRanges()
        If ThreadJobs.Length = nThreads Then Exit Sub
        If iLength = 0 Then Exit Sub
        Dim ThreadBytes As Int64 = iLength \ nThreads
        Array.Resize(ThreadJobs, nThreads)
        For i = 0 To nThreads - 1
            ThreadJobs(i) = New ThreadRangeJob : ThreadJobs(i).Setup(i * ThreadBytes, ((i + 1) * ThreadBytes) - 1, Me)
        Next
        ThreadJobs(nThreads - 1).iEnd = iLength - 1
    End Sub

    Friend Sub ResumeNextThread(ByRef pjob As ThreadRangeJob)
        RaiseEvent ContinueResume(pjob)
    End Sub
    Public Sub SetHandler(ByVal ErrorHander As dsHandleException, ByVal sPath As String)
        JobSavePath = sPath
        HandleException = ErrorHander
        clsDownloader.dParentJob = Me
        InitiateComponents()
    End Sub
    Public Sub SetDownloadDetails(ByVal dURL As String, ByVal dLocation As String, ByVal ErrorHander As dsHandleException, ByVal sPath As String, Optional ByVal Referer As String = "", Optional ByVal cookie As String = "", Optional ByVal Len As Int64 = -1, Optional ByVal cResume As Boolean = False, Optional ByVal Chunked As Boolean = False, Optional ByVal ConnectTimeout As Int64 = 60000, Optional ByVal RecieveTimeout As Int64 = 60000, Optional ByVal SendTimeout As Int64 = 60000, Optional ByVal ConnectRetries As Int64 = 2, Optional ByVal SendRetries As Int64 = 2, Optional ByVal RecieveRetries As Int64 = 2, Optional ByVal BufferSize As Int64 = 4096, Optional ByVal Threads As Int64 = 10, Optional ByVal port As Int64 = 80, Optional ByVal ThreadRetryTime As Int64 = 10000)
        JobSavePath = sPath
        sDownloadURL = dURL
        sFileLocation = dLocation
        HandleException = ErrorHander
        sReferer = Referer
        sCookies = cookie
        canResume = cResume
        isChunked = Chunked
        SetSettings(ConnectTimeout, RecieveTimeout, SendTimeout, ConnectRetries, SendRetries, RecieveRetries, BufferSize, Threads, port, ThreadRetryTime)
        clsDownloader.dParentJob = Me
        iLength = Len
        InitiateComponents()
    End Sub

    Private Sub SetSettings(ByVal ConnectTimeout As Int64, ByVal RecieveTimeout As Int64, ByVal SendTimeout As Int64, ByVal ConnectRetries As Int64, ByVal SendRetries As Int64, ByVal RecieveRetries As Int64, ByVal BufferSize As Int64, ByVal Threads As Int64, ByVal port As Int64, ByVal ThreadRetryTime As Int64)
        iConnectTimeout = ConnectTimeout
        iRecieveTimeout = RecieveTimeout
        iSendTimeout = SendTimeout
        iSendRetries = SendRetries
        iConnectRetries = ConnectRetries
        iRecieveRetries = RecieveRetries
        iBufferSize = BufferSize
        nThreads = Threads
        iPort = port
        iThreadRetryTime = ThreadRetryTime
    End Sub

    Friend Sub UpdateMainStatus(ByVal newStatus As Status)
        If DownloadJobStatus = newStatus Then Exit Sub
        DownloadJobStatus = newStatus
        RaiseEvent JobStatusChanged(newStatus, Me)
    End Sub
    Private Sub InitiateComponents()
        If cStatBoard Is Nothing Then
            cStatBoard = New StatBoard
        End If
        If ThreadJobs Is Nothing Then Array.Resize(ThreadJobs, 1)
        If ThreadJobs(0) Is Nothing Then ThreadJobs(0) = New ThreadRangeJob
        For Each tJob As ThreadRangeJob In ThreadJobs
            If Not tJob Is Nothing Then tJob.dParentJob = Me
        Next
        cStatBoard.Setup(iLength, Me)
        Array.Resize(clsDownloader.DownloadThreads, nThreads)
        Array.Resize(clsDownloader.DownloadSockets, nThreads)
        Array.Resize(cItemLog, nThreads)
        For i = 0 To nThreads - 1
            If Not cItemLog(i) Is Nothing Then
                AddHandler cItemLog(i).ThreadStatusChanged, AddressOf clsDownloader.dParentJob.ThreadStatusChanged
                cItemLog(i).Setup(Me)
                cItemLog(i).Update(InfoLogItem.Direction.Impo, "Thread " & (i + 1) & " Initialized")
                Continue For
            End If
            cItemLog(i) = New InfoLog
            AddHandler cItemLog(i).ThreadStatusChanged, AddressOf clsDownloader.dParentJob.ThreadStatusChanged
            cItemLog(i).UpdateThreadStatus(InfoLog.ThreadState.Queue)
            cItemLog(i).Setup(Me)
            cItemLog(i).Update(InfoLogItem.Direction.Impo, "Thread " & (i + 1) & " Initialized")
        Next
        AddHandler JobStatusChanged, AddressOf cStatBoard.DownloadJobStatusChanged
        If jGUID = Nothing Then
            jGUID = Guid.NewGuid
        End If
        Save()
        ThreadStatusRecheck()
    End Sub

End Class
Public Class ThreadRangeJob
    Public iStart As Int64
    Public iEnd As Int64
    Public iCurrent As Int64
    <XmlIgnore()> Friend dParentJob As DownloadJob
    Friend Sub Setup(ByVal startByte As int64, ByVal endByte As int64, ByVal dJob As DownloadJob, Optional ByVal currentByte As int64 = 0)
        dParentJob = dJob
        iStart = startByte
        iEnd = endByte
        iCurrent = currentByte
    End Sub
    Friend Sub UpdateCurrent(ByVal cur As int64)
        iCurrent = iCurrent + cur
    End Sub
    Friend Function IsClear() As Boolean
        Dim fsThread As New IO.FileStream(dParentJob.sFileLocation, IO.FileMode.OpenOrCreate, IO.FileAccess.ReadWrite, IO.FileShare.ReadWrite)
        Dim buffer(iEnd - iStart) As Byte
        fsThread.Position = iStart
        fsThread.Read(buffer, 0, iEnd - iStart)
        fsThread.Close()
        If buffer.IsNull = True Then Return True Else Return False
    End Function
    Friend Function IsComplete() As Boolean
        If iCurrent = iEnd - iStart + 1 Then Return True
        Return False
    End Function
    Friend Function IsBlank() As Boolean
        If iCurrent = 0 Then Return True
        Return False
    End Function
End Class
Public Class InfoLog
    Enum ThreadState
        Queue = 0
        Starting = 1
        Downloading = 2
        Done = 3
        Stopped = 4
        Paused = 5
    End Enum
    Event ThreadStatusChanged(ByVal newStatus As ThreadState, ByRef iLog As InfoLog)
    Event NewItem(ByVal infoItem As InfoLogItem)
    Public ThreadStatus As ThreadState
    Public InfoItemList As List(Of InfoLogItem)
    <XmlIgnore()> Friend dParentJob As DownloadJob
    Friend Sub UpdateThreadStatus(ByVal newStatus As ThreadState)
        If ThreadStatus = newStatus Then Exit Sub
        ThreadStatus = newStatus
        RaiseEvent ThreadStatusChanged(newStatus, Me)
    End Sub
    Friend Sub Setup(ByRef dJob As DownloadJob)
        InfoItemList = New List(Of InfoLogItem)
        dParentJob = dJob
        If ThreadStatus = ThreadState.Downloading Then UpdateThreadStatus(ThreadState.Queue)
        If ThreadStatus = ThreadState.Starting Then UpdateThreadStatus(ThreadState.Queue)
        UpdateThreadStatus(ThreadStatus)
    End Sub
    Friend Sub Update(ByVal eDirection As InfoLogItem.Direction, ByVal sText As String)
        Dim nItem As New InfoLogItem
        nItem.Setup(eDirection, sText)
        InfoItemList.Add(nItem)
        RaiseEvent NewItem(nItem)
    End Sub
    Private Sub Clear()
        InfoItemList.Clear()
    End Sub
End Class
Public Class InfoLogItem
    Enum Direction
        To_Server = 0
        To_Client = 1
        _Error = 2
        Link = 3
        Impo = 4
    End Enum
    Public ItemColor As Drawing.Color
    Public ItemDirection As Direction
    Public ItemText As String
    Public ItemTime As String
    Sub Setup(ByVal eDirection As Direction, ByVal sText As String)
        ItemText = sText
        ItemDirection = eDirection
        If ItemDirection = Direction.To_Client Then ItemColor = Drawing.Color.PaleGreen
        If ItemDirection = Direction.To_Server Then ItemColor = Drawing.Color.LightCyan
        If ItemDirection = Direction._Error Then ItemColor = Drawing.Color.MistyRose
        If ItemDirection = Direction.Link Then ItemColor = Drawing.Color.LightSkyBlue
        If ItemDirection = Direction.Impo Then ItemColor = Drawing.Color.LightSkyBlue
        ItemTime = Now.TimeOfDay.ToString.Substring(0, 8)
    End Sub
End Class
Public Class StatBoard
    Dim startTime As TimeSpan
    Public sDuration As String
    Public sETA As String
    Public iDownloaded As int64
    Public iTotal As int64
    Public iRemaining As int64
    Public dSpeedAvg As int64
    Public iSpeed As int64
    Public dPercent As Decimal
    Dim savedRunTime As TimeSpan
    Dim currentRunTime As TimeSpan
    Dim requiredTime As TimeSpan
    <XmlIgnore()> Friend dParentJob As DownloadJob
    Friend Sub Setup(ByVal iBytesTotal As int64, ByVal dJob As DownloadJob)
        dParentJob = dJob
        iTotal = iBytesTotal
    End Sub
    Friend Sub DownloadJobStatusChanged(ByVal newStatus As DownloadJob.Status, ByRef downJob As DownloadJob)
        If newStatus = DownloadJob.Status.Downloading Then StartTiming()
        If newStatus = DownloadJob.Status.Stopped Or newStatus = DownloadJob.Status.Queue Or newStatus = DownloadJob.Status.Done Then StopTiming()
    End Sub
    Private Sub StartTiming()
        If sDuration <> "" Then savedRunTime = New TimeSpan(sDuration.Substring(0, 2), sDuration.Substring(3, 2), sDuration.Substring(6, 2))
        startTime = Now.TimeOfDay
        Crunch()
    End Sub
    Private Sub StopTiming()
        Crunch()
    End Sub
    Friend Sub AddDownloaded(ByVal nBytes As int64)
        iSpeed = iSpeed + nBytes
        iDownloaded = iDownloaded + nBytes
        Crunch()
    End Sub
    Private Sub Crunch()
        currentRunTime = savedRunTime.Add(Now.TimeOfDay.Subtract(startTime))
        sDuration = currentRunTime.ToString.Substring(0, 8)
        If iTotal <= 0 Then Exit Sub
        If iDownloaded = iTotal Then dParentJob.UpdateMainStatus(DownloadJob.Status.Done)
        dPercent = (iDownloaded * 100) / iTotal
        iRemaining = iTotal - iDownloaded
        If currentRunTime.TotalSeconds <= 0 Then Exit Sub
        dSpeedAvg = iDownloaded / currentRunTime.TotalSeconds
        If dSpeedAvg > 0 Then requiredTime = New TimeSpan(0, 0, iRemaining / dSpeedAvg) Else requiredTime = New TimeSpan(99, 99, 99)
        sETA = requiredTime.ToString.Substring(0, 8)
    End Sub
    Friend Sub New()
        If sETA = "" Then sETA = "00:00:00"
        If sDuration = "" Then sDuration = "00:00:00"
        If savedRunTime = Nothing Then savedRunTime = New TimeSpan(0, 0, 0)
        If currentRunTime = Nothing Then currentRunTime = New TimeSpan(0, 0, 0)
        If requiredTime = Nothing Then requiredTime = New TimeSpan(0, 0, 0)
    End Sub
End Class
